@@ -1,9 +1,37 @@
+/* $EPIC: exec.c,v 1.24 2003/12/04 22:59:03 crazyed Exp $ */
 /*
  * exec.c: handles exec'd process for IRCII 
  *
- * Copyright 1990 Michael Sandrof
- * Copyright 1997 EPIC Software Labs
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1997, 2003 EPIC Software Labs
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 
 #include "irc.h"
@@ -64,7 +92,8 @@ static	int	process_list_size = 0;
 
 static 	void 	handle_filedesc 	(Process *, int *, int, int);
 static 	void 	cleanup_dead_processes 	(void);
-static 	void 	ignore_process 		(int index);
+static 	void 	ignore_process 		(int);
+static 	void 	close_in		(int);
 static 	void 	kill_process 		(int, int);
 static 	void 	kill_all_processes 	(int signo);
 static 	int 	valid_process_index 	(int proccess);
@@ -88,13 +117,13 @@ static	int 	logical_to_index 	(const char *logical);
  */
 BUILT_IN_COMMAND(execcmd)
 {
-	char		*who = NULL,
-			*logical = NULL,
-			*redirect = NULL,
+	const char	*who = NULL;
+	char		*logical = NULL,
 			*flag;
+	const char 	*redirect = NULL;
 	unsigned 	refnum = 0;
+	size_t		len;
 	int		sig,
-			len,
 			i,
 			refnum_flag = 0,
 			logical_flag = 0;
@@ -111,8 +140,6 @@ BUILT_IN_COMMAND(execcmd)
 	 */
 	if (!*args)
 	{
-		int	i;
-
 		if (!process_list)
 		{
 			say("No processes are running");
@@ -122,8 +149,6 @@ BUILT_IN_COMMAND(execcmd)
 		say("Process List:");
 		for (i = 0; i < process_list_size; i++)
 		{
-			Process	*proc;
-
 			if (((proc = process_list[i])) == NULL)
 				continue;
 
@@ -151,12 +176,12 @@ BUILT_IN_COMMAND(execcmd)
 		 */
 		if (my_strnicmp(flag, "OUT", len) == 0)
 		{
-			if (doing_privmsg)
+			if (get_server_doing_privmsg(from_server))
 				redirect = "NOTICE";
 			else
 				redirect = "PRIVMSG";
 
-			if (!(who = get_channel_by_refnum(0)))
+			if (!(who = get_echannel_by_refnum(0)))
 			{
 			     say("No current channel in this window for -OUT");
 			     return;
@@ -193,7 +218,7 @@ BUILT_IN_COMMAND(execcmd)
 		 */
 		else if (my_strnicmp(flag, "MSG", len) == 0)
 		{
-			if (doing_privmsg)
+			if (get_server_doing_privmsg(from_server))
 				redirect = "NOTICE";
 			else
 				redirect = "PRIVMSG";
@@ -263,6 +288,19 @@ BUILT_IN_COMMAND(execcmd)
 			return;
 		}
 
+		/*
+		 * /EXEC -CLOSEIN close the processes STDIN,
+		 * in the hope that it will take the hint.
+		 */
+		else if (my_strnicmp(flag, "CLOSEIN", len) == 0)
+		{
+			if ((i = get_process_index(&args)) == -1)
+				return;
+
+			close_in(i);
+			return;
+		}
+
 		/*
 		 * /EXEC -NOTICE <target> redirects the output of an /exec 
 		 * to a specified target
@@ -356,8 +394,6 @@ BUILT_IN_COMMAND(execcmd)
 	 */
 	if (*args == '%')
 	{
-		int i;
-
 		/*
 		 * Make sure the process is actually running
 		 */
@@ -404,7 +440,20 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			else 
 				say("The name %s is not unique!", logical);
 		}
-		message_to(0);
+
+		/*
+		 * Change the stdout and stderr lines _if_ they are given.
+		 */
+		if (stdoutc)
+			malloc_strcpy(&proc->stdoutc, stdoutc);
+		if (stdoutpc)
+			malloc_strcpy(&proc->stdoutpc, stdoutpc);
+		if (stderrc)
+			malloc_strcpy(&proc->stderrc, stderrc);
+		if (stderrpc)
+			malloc_strcpy(&proc->stderrpc, stderrpc);
+
+		message_to(-1);
 	}
 
 	/*
@@ -415,7 +464,6 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 		int	p0[2], p1[2], p2[2],
 			pid, cnt;
 		char	*shell,
-			*flag,
 			*arg;
 		char	*name = args;
 
@@ -454,8 +502,6 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 		 */
 		case 0:
 		{
-			int i;
-
 			/*
 			 * Fire up a new job control session,
 			 * Sever all ties we had with the parent ircII process
@@ -477,7 +523,7 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			/*
 			 * Pretend to be just a dumb terminal
 			 */
-			bsd_setenv("TERM", "tty", 1);
+			setenv("TERM", "tty", 1);
 
 			/*
 			 * Figure out what shell (if any) we're using
@@ -490,24 +536,24 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			 */
 			if (direct || !shell)
 			{
-				char	**args;
 				int	max;
+				char **my_args;
 
 				cnt = 0;
 				max = 5;
-				args = new_malloc(sizeof(char *) * max);
+				my_args = new_malloc(sizeof(char *) * max);
 				while ((arg = new_next_arg(name, &name)))
 				{
-					if (cnt == max)
+					my_args[cnt] = arg;
+
+					if (++cnt >= max)
 					{
 						max += 5;
-						RESIZE(args, char *, max);
+						RESIZE(my_args, char *, max);
 					}
-
-					args[cnt++] = arg;
 				}
-				args[cnt] = NULL;
-				execvp(args[0], args);
+				my_args[cnt] = NULL;
+				execvp(my_args[0], my_args);
 			}
 
 			/*
@@ -535,9 +581,7 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 		 */
 		default:
 		{
-			int	i;
-			Process	*proc = new_malloc(sizeof(Process));
-
+			proc = new_malloc(sizeof(Process));
 			new_close(p0[0]);
 			new_close(p1[1]);
 			new_close(p2[1]);
@@ -577,9 +621,9 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			/*
 			 * Fill in the new proc entry
 			 */
-			proc->name = m_strdup(name);
+			proc->name = malloc_strdup(name);
 			if (logical)
-				proc->logical = m_strdup(logical);
+				proc->logical = malloc_strdup(logical);
 			else
 				proc->logical = NULL;
 			proc->index = i;
@@ -590,8 +634,8 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			proc->refnum = refnum;
 			proc->redirect = NULL;
 			if (redirect)
-				proc->redirect = m_strdup(redirect);
-			proc->server = current_window->server;
+				proc->redirect = malloc_strdup(redirect);
+			proc->server = from_server;
 			proc->counter = 0;
 			proc->exited = 0;
 			proc->termsig = 0;
@@ -600,26 +644,26 @@ say("Output from process %d (%s) now going to you", i, proc->name);
 			proc->who = NULL;
 			proc->disowned = 0;
 			if (who)
-				proc->who = m_strdup(who);
+				proc->who = malloc_strdup(who);
 			proc->dumb = 0;
 
 			if (stdoutc)
-				proc->stdoutc = m_strdup(stdoutc);
+				proc->stdoutc = malloc_strdup(stdoutc);
 			else
 				proc->stdoutc = NULL;
 
 			if (stdoutpc)
-				proc->stdoutpc = m_strdup(stdoutpc);
+				proc->stdoutpc = malloc_strdup(stdoutpc);
 			else
 				proc->stdoutpc = NULL;
 
 			if (stderrc)
-				proc->stderrc = m_strdup(stderrc);
+				proc->stderrc = malloc_strdup(stderrc);
 			else
 				proc->stderrc = NULL;
 
 			if (stderrpc)
-				proc->stderrpc = m_strdup(stderrpc);
+				proc->stderrpc = malloc_strdup(stderrpc);
 			else
 				proc->stderrpc = NULL;
 
@@ -643,7 +687,7 @@ say("Output from process %d (%s) now going to you", i, proc->name);
  * are closed.  If EOF has been asserted on both, then  we mark the process
  * as being "dumb".  Once it is reaped (exited), it is expunged.
  */
-void 		do_processes (fd_set *rd)
+void 		do_processes (fd_set *rd, fd_set *wd)
 {
 	int	i;
 	int	limit;
@@ -687,9 +731,9 @@ void 		do_processes (fd_set *rd)
 static void 	handle_filedesc (Process *proc, int *fd, int hook_nonl, int hook_nl)
 {
 	char 	exec_buffer[IO_BUFFER_SIZE + 1];
-	int	len;
+	ssize_t	len;
 
-	switch ((len = dgets(exec_buffer, *fd, 0)))	/* No buffering! */
+	switch ((len = dgets(*fd, exec_buffer, IO_BUFFER_SIZE, 0, NULL))) /* No buffering! */
 	{
 		case -1:	/* Something died */
 		{
@@ -712,10 +756,10 @@ static void 	handle_filedesc (Process *proc, int *fd, int hook_nonl, int hook_nl
 				goto this_sucks;
 			}
 
-			     if (hook_nl == EXEC_LIST && proc->stdoutpc)
+			     if (hook_nl == EXEC_LIST && proc->stdoutpc && *proc->stdoutpc)
 				parse_line("EXEC", proc->stdoutpc, 
 						exec_buffer, 0, 0);
-			else if (hook_nl == EXEC_ERRORS_LIST && proc->stderrpc)
+			else if (hook_nl == EXEC_ERRORS_LIST && proc->stderrpc && *proc->stderrpc)
 				parse_line("EXEC", proc->stderrpc, 
 						exec_buffer, 0, 0);
 			else if (proc->logical)
@@ -726,12 +770,16 @@ static void 	handle_filedesc (Process *proc, int *fd, int hook_nonl, int hook_nl
 					proc->index, exec_buffer);
 
 			set_prompt_by_refnum(proc->refnum, exec_buffer);
-			update_input(UPDATE_ALL);
 			break;
 		}
 		default:	/* We got a full line */
-		{
 this_sucks:
+		{
+			int ofs;
+
+			ofs = from_server;
+			if (proc->refnum)
+				from_server = proc->server;
 			message_to(proc->refnum);
 			proc->counter++;
 
@@ -745,10 +793,10 @@ this_sucks:
 				redirect_text(proc->server, proc->who, 
 					exec_buffer, proc->redirect, 1);
 
-			     if (hook_nl == EXEC_LIST && proc->stdoutc)
+			if (hook_nl == EXEC_LIST && proc->stdoutc && *proc->stdoutc)
 				parse_line("EXEC", proc->stdoutc, 
 						exec_buffer, 0, 0);
-			else if (hook_nl == EXEC_ERRORS_LIST && proc->stderrc)
+			else if (hook_nl == EXEC_ERRORS_LIST && proc->stderrc && *proc->stderrc)
 				parse_line("EXEC", proc->stderrc, 
 						exec_buffer, 0, 0);
 			else if (proc->logical)
@@ -766,7 +814,8 @@ this_sucks:
 						put_it("%s", exec_buffer);
 			}
 
-			message_to(0);
+			message_to(-1);
+			from_server = ofs;
 		}
 	}
 }
@@ -775,7 +824,7 @@ this_sucks:
  * get_child_exit: This looks for dead child processes of the client.
  * There are two main sources of dead children:  Either an /exec'd process
  * has exited, or the client has attempted to fork() off a helper process
- * (such as wserv or gzip) and that process has choked on itself.  
+ * (such as wserv or gzip) and that process has choked on itself.
  *
  * When SIGCHLD is recieved, the global variable 'dead_children_processes'
  * is incremented.  When this function is called, we go through and call
@@ -797,13 +846,10 @@ int 		get_child_exit (pid_t wanted)
 	 * or we've found the one asked for.
 	 */
 	if (dead_children_processes)
-	while ((pid = waitpid(wanted, &status, WNOHANG)) > 0)
 	{
-		/*
-		 * Ideally, this should never be < 0.
-		 */
-		dead_children_processes--;
-
+	    block_signal(SIGCHLD);
+	    while ((pid = waitpid(wanted, &status, WNOHANG)) > 0)
+	    {
 		/*
 		 * First thing we do is look to see if the process we're
 		 * working on is the one that was asked for.  If it is,
@@ -811,9 +857,9 @@ int 		get_child_exit (pid_t wanted)
 		 */
 		if (wanted != -1 && pid == wanted)
 		{
-			if (WIFEXITED(status))   
+			if (WIFEXITED(status))
 				return WEXITSTATUS(status);
-			if (WIFSTOPPED(status))  
+			if (WIFSTOPPED(status))
 				return -(WSTOPSIG(status));
 			if (WIFSIGNALED(status)) 
 				return -(WTERMSIG(status));
@@ -838,6 +884,9 @@ int 		get_child_exit (pid_t wanted)
 				}
 			}
 		}
+	    }
+	    dead_children_processes = 0;
+	    unblock_signal(SIGCHLD);
 	}
 
 	/*
@@ -869,6 +918,7 @@ void 		clean_up_processes (void)
 		say("Closing all left over exec's");
 		kill_all_processes(SIGTERM);
 		sleep(2);
+		get_child_exit(-1);
 		kill_all_processes(SIGKILL);
 	}
 }
@@ -884,6 +934,7 @@ int 		text_to_process (int proc_index, const char *text, int show)
 {
 	Process	*	proc;
 	char	*	my_buffer;
+	size_t	size;
 
 	if (valid_process_index(proc_index) == 0)
 		return 1;
@@ -893,31 +944,17 @@ int 		text_to_process (int proc_index, const char *text, int show)
 	message_to(proc->refnum);
 	if (show)
 		put_it("%s%s", get_prompt_by_refnum(proc->refnum), text);
-	message_to(0);
+	message_to(-1);
 
-	my_buffer = alloca(strlen(text) + 2);
-	/* Obviously this is safe. */
-	strcpy(my_buffer, text);
-	strcat(my_buffer, "\n");
+	size = strlen(text) + 2;
+	my_buffer = alloca(size);
+	snprintf(my_buffer, size, "%s\n", text);
 	write(proc->p_stdin, my_buffer, strlen(my_buffer));
 	set_prompt_by_refnum(proc->refnum, empty_string);
 
 	return (0);
 }
 
-/*
- * When a server goes away, re-assign all of the /exec's that are
- * current bound to that server.
- */
-void 		exec_server_delete (int i)
-{
-	int	j;
-
-	for (j = 0; j < process_list_size; j++)
-		if (process_list[j] && process_list[j]->server >= i)
-			process_list[j]->server--;
-}
-
 /*
  * This adds a new /wait %proc -cmd   entry to a running process.
  */
@@ -928,7 +965,7 @@ void 		add_process_wait (int proc_index, const char *cmd)
 
 	new_ewl = new_malloc(sizeof(List));
 	new_ewl->next = NULL;
-	new_ewl->name = m_strdup(cmd);
+	new_ewl->name = malloc_strdup(cmd);
 
 	if ((posn = proc->waitcmds))
 	{
@@ -977,7 +1014,7 @@ static void 	cleanup_dead_processes (void)
 	int	i;
 	List	*cmd,
 		*next;
-	Process *dead, *proc;
+	Process *deadproc, *proc;
 	char	*exit_info;
 
 	if (!process_list)
@@ -996,34 +1033,34 @@ static void 	cleanup_dead_processes (void)
 		if ((!proc->exited || !proc->dumb) && !proc->disowned)
 			continue;		/* Not really dead yet */
 
-		dead = process_list[i];
+		deadproc = process_list[i];
 		process_list[i] = NULL;
 
 		/*
 		 * First thing to do is fill out the exit information
 		 */
-		if (dead->logical)
+		if (deadproc->logical)
 		{
-			size_t	len = strlen(dead->logical) + 25;
+			size_t	len = strlen(deadproc->logical) + 25;
 
 			exit_info = alloca(len);
 			snprintf(exit_info, len, "%s %d %d", 
-					dead->logical, dead->termsig,
-					dead->retcode);
+					deadproc->logical, deadproc->termsig,
+					deadproc->retcode);
 		}
 		else
 		{
 			exit_info = alloca(40);
 			snprintf(exit_info, 32, "%d %d %d",
-				dead->index, dead->termsig, dead->retcode);
+				deadproc->index, deadproc->termsig, deadproc->retcode);
 		}
 
 
 		/*
 		 * First thing we do is run any /wait %proc -cmd commands
 		 */
-		next = dead->waitcmds;
-		dead->waitcmds = NULL;
+		next = deadproc->waitcmds;
+		deadproc->waitcmds = NULL;
 		while ((cmd = next))
 		{
 			next = cmd->next;
@@ -1039,26 +1076,30 @@ static void 	cleanup_dead_processes (void)
 		{
 			if (get_int_var(NOTIFY_ON_TERMINATION_VAR))
 			{
-				if (dead->termsig > 0 && dead->termsig < NSIG)
+				if (deadproc->termsig > 0 && deadproc->termsig < NSIG)
 	say("Process %d (%s) terminated with signal %s (%d)", 
-	   dead->index, dead->name, sys_siglist[dead->termsig], dead->termsig);
-				else if (dead->disowned)
-	say("Process %d (%s) disowned", dead->index, dead->name);
+	   deadproc->index, deadproc->name, sys_siglist[deadproc->termsig], deadproc->termsig);
+				else if (deadproc->disowned)
+	say("Process %d (%s) disowned", deadproc->index, deadproc->name);
 				else
 	say("Process %d (%s) terminated with return code %d", 
-		dead->index, dead->name, dead->retcode);
+		deadproc->index, deadproc->name, deadproc->retcode);
 			}
 		}
 
-		if (dead->p_stdin != -1)
-			close(dead->p_stdin);
-		new_close(dead->p_stdout);
-		new_close(dead->p_stderr);
-		new_free(&dead->name);
-		new_free(&dead->logical);
-		new_free(&dead->who);
-		new_free(&dead->redirect);
-		new_free((char **)&dead);
+		if (deadproc->p_stdin != -1)
+			new_close(deadproc->p_stdin);
+		new_close(deadproc->p_stdout);
+		new_close(deadproc->p_stderr);
+		new_free(&deadproc->name);
+		new_free(&deadproc->logical);
+		new_free(&deadproc->who);
+		new_free(&deadproc->redirect);
+		new_free(&deadproc->stdoutc);
+		new_free(&deadproc->stdoutpc);
+		new_free(&deadproc->stderrc);
+		new_free(&deadproc->stderrpc);
+		new_free((char **)&deadproc);
 	}
 
 	/*
@@ -1086,14 +1127,14 @@ static void 	cleanup_dead_processes (void)
  * condition on its output fd, so it will probably either take the hint, or
  * its output will go the bit bucket (which we want to happen)
  */
-static void 	ignore_process (int index)
+static void 	ignore_process (int idx)
 {
 	Process *proc;
 
-	if (valid_process_index(index) == 0)
+	if (valid_process_index(idx) == 0)
 		return;
 
-	proc = process_list[index];
+	proc = process_list[idx];
 	if (proc->p_stdin != -1)
 		proc->p_stdin = new_close(proc->p_stdin);
 	if (proc->p_stdout != -1)
@@ -1104,6 +1145,23 @@ static void 	ignore_process (int index)
 	proc->dumb = 1;
 }
 
+/*
+ * close_in:  When we are finished with the process but still want the
+ * rest of its output, we close its input, and hopefully it will get the
+ * message and close up shop.
+ */
+static void 	close_in (int idx)
+{
+	Process *proc;
+
+	if (valid_process_index(idx) == 0)
+		return;
+
+	proc = process_list[idx];
+	if (proc->p_stdin != -1)
+		proc->p_stdin = new_close(proc->p_stdin);
+}
+
 
 
 
